/*
 *  libpipi       Proper image processing implementation library
 *  Copyright (c) 2004-2008 Sam Hocevar <sam@zoy.org>
 *                All Rights Reserved
 *
 *  $Id$
 *
 *  This library is free software. It comes without any warranty, to
 *  the extent permitted by applicable law. You can redistribute it
 *  and/or modify it under the terms of the Do What The Fuck You Want
 *  To Public License, Version 2, as published by Sam Hocevar. See
 *  http://sam.zoy.org/wtfpl/COPYING for more details.
 */

/*
 * convolution_template.h: convolution templates
 * required macros:
 *   FUNC1  standard function name
 *   FUNC2  separable function name
 *   PIXEL  pixel type
 *   FLAGS  1 (grayscale)
 *          2 (loop)
 */

static pipi_image_t *FUNC1(pipi_image_t *src, int m, int n, double mat[]);
static pipi_image_t *FUNC2(pipi_image_t *src, int m, double hvec[],
                                              int n, double vvec[]);

static pipi_image_t *FUNC1(pipi_image_t *src, int m, int n, double mat[])
{
    pipi_image_t *dst;
    pipi_pixels_t *srcp, *dstp;
    PIXEL *srcdata, *dstdata;
    int x, y, i, j, w, h;

    w = src->w;
    h = src->h;

    srcp = FLAG_GRAY ? pipi_getpixels(src, PIPI_PIXELS_Y_F)
                     : pipi_getpixels(src, PIPI_PIXELS_RGBA_F);
    srcdata = (PIXEL *)srcp->pixels;

    dst = pipi_new(w, h);
    dstp = FLAG_GRAY ? pipi_getpixels(dst, PIPI_PIXELS_Y_F)
                     : pipi_getpixels(dst, PIPI_PIXELS_RGBA_F);
    dstdata = (PIXEL *)dstp->pixels;

    for(y = 0; y < h; y++)
    {
        for(x = 0; x < w; x++)
        {
            if(FLAG_GRAY)
            {
                double Y = 0.;
                int x2, y2;

                for(j = 0; j < n; j++)
                {
                    y2 = y + j - n / 2;
                    if(y2 < 0) y2 = FLAG_WRAP ? (h - 1 - (-y2 - 1) % h) : 0;
                    else if(y2 >= h) y2 = FLAG_WRAP ? y2 % h : h - 1;

                    for(i = 0; i < m; i++)
                    {
                        x2 = x + i - m / 2;
                        if(x2 < 0) x2 = FLAG_WRAP ? (w - 1 - (-x2 - 1) % w) : 0;
                        else if(x2 >= w) x2 = FLAG_WRAP ? x2 % w : w - 1;

                        Y += mat[j * m + i] * srcdata[y2 * w + x2];
                    }
                }

                dstdata[y * w + x] = Y;
            }
            else
            {
                double R = 0., G = 0., B = 0.;
                int x2, y2, off = 4 * (y * w + x);

                for(j = 0; j < n; j++)
                {
                    y2 = y + j - n / 2;
                    if(y2 < 0) y2 = FLAG_WRAP ? (h - 1 - (-y2 - 1) % h) : 0;
                    else if(y2 >= h) y2 = FLAG_WRAP ? y2 % h : h - 1;

                    for(i = 0; i < m; i++)
                    {
                        double f = mat[j * m + i];

                        x2 = x + i - m / 2;
                        if(x2 < 0) x2 = FLAG_WRAP ? (w - 1 - (-x2 - 1) % w) : 0;
                        else if(x2 >= w) x2 = FLAG_WRAP ? x2 % w : w - 1;

                        R += f * srcdata[(y2 * w + x2) * 4];
                        G += f * srcdata[(y2 * w + x2) * 4 + 1];
                        B += f * srcdata[(y2 * w + x2) * 4 + 2];
                    }
                }

                dstdata[off] = R;
                dstdata[off + 1] = G;
                dstdata[off + 2] = B;
            }
        }
    }

    return dst;
}

static pipi_image_t *FUNC2(pipi_image_t *src,
                           int m, double hvec[], int n, double vvec[])
{
    pipi_image_t *dst;
    pipi_pixels_t *srcp, *dstp;
    PIXEL *srcdata, *dstdata;
    double *buffer;
    int x, y, i, j, w, h;

    w = src->w;
    h = src->h;

    srcp = FLAG_GRAY ? pipi_getpixels(src, PIPI_PIXELS_Y_F)
                     : pipi_getpixels(src, PIPI_PIXELS_RGBA_F);
    srcdata = (PIXEL *)srcp->pixels;

    dst = pipi_new(w, h);
    dstp = FLAG_GRAY ? pipi_getpixels(dst, PIPI_PIXELS_Y_F)
                     : pipi_getpixels(dst, PIPI_PIXELS_RGBA_F);
    dstdata = (PIXEL *)dstp->pixels;

    buffer = malloc(w * h * (FLAG_GRAY ? 1 : 4) * sizeof(double));

    for(y = 0; y < h; y++)
    {
        for(x = 0; x < w; x++)
        {
            if(FLAG_GRAY)
            {
                double Y = 0.;
                int x2;

                for(i = 0; i < m; i++)
                {
                    x2 = x + i - m / 2;
                    if(x2 < 0) x2 = FLAG_WRAP ? (w - 1 - (-x2 - 1) % w) : 0;
                    else if(x2 >= w) x2 = FLAG_WRAP ? x2 % w : w - 1;

                    Y += hvec[i] * srcdata[y * w + x2];
                }

                buffer[y * w + x] = Y;
            }
            else
            {
                double R = 0., G = 0., B = 0.;
                int x2, off = 4 * (y * w + x);

                for(i = 0; i < m; i++)
                {
                    double f = hvec[i];

                    x2 = x + i - m / 2;
                    if(x2 < 0) x2 = FLAG_WRAP ? (w - 1 - (-x2 - 1) % w) : 0;
                    else if(x2 >= w) x2 = FLAG_WRAP ? x2 % w : w - 1;

                    R += f * srcdata[(y * w + x2) * 4];
                    G += f * srcdata[(y * w + x2) * 4 + 1];
                    B += f * srcdata[(y * w + x2) * 4 + 2];
                }

                buffer[off] = R;
                buffer[off + 1] = G;
                buffer[off + 2] = B;
            }
        }
    }

    for(y = 0; y < h; y++)
    {
        for(x = 0; x < w; x++)
        {
            if(FLAG_GRAY)
            {
                double Y = 0.;
                int y2;

                for(j = 0; j < n; j++)
                {
                    y2 = y + j - n / 2;
                    if(y2 < 0) y2 = FLAG_WRAP ? (h - 1 - (-y2 - 1) % h) : 0;
                    else if(y2 >= h) y2 = FLAG_WRAP ? y2 % h : h - 1;

                    Y += vvec[j] * buffer[y2 * w + x];
                }

                dstdata[y * w + x] = Y;
            }
            else
            {
                double R = 0., G = 0., B = 0.;
                int y2, off = 4 * (y * w + x);

                for(j = 0; j < n; j++)
                {
                    double f = vvec[j];

                    y2 = y + j - n / 2;
                    if(y2 < 0) y2 = FLAG_WRAP ? (h - 1 - (-y2 - 1) % h) : 0;
                    else if(y2 >= h) y2 = FLAG_WRAP ? y2 % h : h - 1;

                    R += f * buffer[(y2 * w + x) * 4];
                    G += f * buffer[(y2 * w + x) * 4 + 1];
                    B += f * buffer[(y2 * w + x) * 4 + 2];
                }

                dstdata[off] = R;
                dstdata[off + 1] = G;
                dstdata[off + 2] = B;
            }
        }
    }

    free(buffer);

    return dst;
}

